{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC                                                                             ##}
{##                                                                                 ##}
{## This is the mainfile of the program, where the execution starts...              ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


Program ABC;


{$APPTYPE CONSOLE}
{$I ABC_Include.pas}


Uses
  Classes,
  SysUtils,
  ABC_Global_Unit in 'ABC_Global_Unit.pas',
  ABC_IO_Unit in 'ABC_IO_Unit.pas',
  ABC_MCP_Unit in 'ABC_MCP_Unit.pas',
  ABC_BWT_Unit in 'ABC_BWT_Unit.pas',
  ABC_CRC_Unit in 'ABC_CRC_Unit.pas',
  ABC_MTF_Unit in 'ABC_MTF_Unit.pas',
  ABC_Model_Unit in 'ABC_Model_Unit.pas',
  ABC_Coder_Unit in 'ABC_Coder_Unit.pas',
  ABC_RLE_EXP_Unit in 'ABC_RLE_EXP_Unit.pas',
  ABC_SIF_Unit in 'ABC_SIF_Unit.pas',
  ABC_AWFC_Unit in 'ABC_AWFC_Unit.pas',
  ABC_RLE_BIT_Unit in 'ABC_RLE_BIT_Unit.pas';

Var
  ABC_IO                    : T_ABC_IO;
  Commands                  : T_R_ABC_Commands;
  I                         : Integer;
  File_List                 : TStringList;
  Input_Filename            : String;
  Output_Filename           : String;
  S                         : String;
  Input_Size                : Double;
  Output_Size               : Double;
  Compression_Rate          : Double;
  Starttime                 : TTimeStamp;
  Endtime                   : TTimeStamp;
  Execution_Time            : Extended;
  Total_Starttime           : TTimeStamp;
  Total_Endtime             : TTimeStamp;
  File_N                    : Double;
  Average_Compression_Rate  : Double;
  Total_Input_Size          : Double;
  Total_Output_Size         : Double;

Begin
  { Get starttime }
  Total_Starttime           := DateTimeToTimeStamp (Now);
  Average_Compression_Rate  := 0.0;
  Total_Input_Size          := 0.0;
  Total_Output_Size         := 0.0;
  File_N                    := 0;

  { Print start message }
  For I := 0 to ABC_IO_START_TEXT_N - 1 do
    Begin { For }
      WriteLn (ABC_A_IO_START_TEXT [I]);
    End; { For }

  ABC_IO := T_ABC_IO.Create;
  Try
    ABC_IO.Parse_Command_Line (Commands);

    { HELP }
    If Commands.Help_F = TRUE then
      Begin { then }
        { Print help message }
        For I := 0 to ABC_IO_HELP_TEXT_N - 1 do
          Begin { For }
            WriteLn (ABC_A_IO_HELP_TEXT [I]);
          End; { For }
      End; { then }

    { COMPRESS }
    If Commands.Compress_F = TRUE then
      Begin { then }
        { Compress all files }
        File_List := TStringList.Create;
        Try
          ABC_IO.Get_Filenames (Commands.Input_Filename, File_List);
          File_N  := File_List.Count;

          If File_N > 0 then
            Begin { then }
              If Commands.Verbose_F = FALSE then
                Begin { then }
                  Write (Format (ABC_IO_COMPRESS_START_TEXT, [File_N]));
                End; { then }

              ABC_MCP := T_ABC_MCP.Create;
              Try
                For I := 0 to File_List.Count - 1 do
                  Begin { For }
                    { Get input filename and output filename }
                    Input_Filename  := File_List.Strings [I];
                    If ExtractFileExt (Input_Filename) <> '' then
                      Begin { then }
                        Output_Filename := Copy (Input_Filename, 1, Pos (ExtractFileExt (Input_Filename), Input_Filename)) + Commands.Output_Extension;
                      End { then }
                        else
                      Begin { else }
                        Output_Filename := Input_Filename + ABC_IO_DELIMITER + Commands.Output_Extension;
                      End; { else }
                    If Input_Filename = Output_Filename then
                      Begin { then }
                        Output_Filename := Output_Filename + ABC_IO_DELIMITER + Commands.Output_Extension
                      End; { then }

                    { Print command }
                    S := ABC_IO_COMPRESS_TEXT + Input_Filename + ABC_IO_BETWEEN_TEXT + Output_Filename;
                    If Commands.Verbose_F = TRUE then
                      Begin { then }
                        WriteLn (S);
                      End { then }
                        else
                      Begin { else }
                        Write (ABC_A_IO_WAIT_TEXT [I mod ABC_IO_WAIT_TEXT_N]);
                      End; { else }

                    { Compress }
                    Starttime := DateTimeToTimeStamp (Now);
                    ABC_MCP.Compress_File (Input_Filename, Output_Filename, Commands.Fast_F);
                    Endtime := DateTimeToTimeStamp (Now);

                    { Print stages }
                    If ABC_MCP.Compress_Information.Fast_F = TRUE then
                      Begin { then }
                        S := ABC_IO_MTF_COMPRESS_TEXT;
                      End { then }
                        else
                      Begin { else }
                        If ABC_MCP.Compress_Information.SIF_F = TRUE then
                          Begin { then }
                            S := ABC_IO_SIF_COMPRESS_TEXT;
                          End { then }
                            else
                          Begin { else }
                            S := ABC_IO_AWFC_COMPRESS_TEXT;
                          End; { else }
                      End; { else }
                    If Commands.Verbose_F = TRUE then
                      Begin { then }
                        WriteLn (S);
                      End; { then }

                    { Print input size and output size }
                    Input_Size        := ABC_IO.Get_File_Size (Input_Filename);
                    Total_Input_Size  := Total_Input_Size + Input_Size;
                    Output_Size       := ABC_IO.Get_File_Size (Output_Filename);
                    Total_Output_Size := Total_Output_Size + Output_Size;
                    If Commands.Verbose_F = TRUE then
                      Begin { then }
                        WriteLn (Format (ABC_IO_INPUT_SIZE_TEXT, [Input_Size]));
                        WriteLn (Format (ABC_IO_OUTPUT_SIZE_TEXT, [Output_Size]));
                      End; { then }

                    { Print compression rate }
                    If Input_Size <> 0 then
                      Begin { then }
                        { Don't devide by zero }
                        Compression_Rate := Output_Size / Input_Size * ABC_BYTE_BIT_SIZE;
                        If Commands.Verbose_F = TRUE then
                          Begin { then }
                            WriteLn (Format (ABC_IO_COMPRESSION_RATIO_TEXT, [Compression_Rate]));
                          End; { then }
                        Average_Compression_Rate := Average_Compression_Rate + Compression_Rate;
                      End; { then }

                    { Print compress time }
                    Execution_Time  := (TimeStampToMSecs (Endtime) - TimeStampToMSecs (Starttime)) / 1000;
                    If Commands.Verbose_F = TRUE then
                      Begin { then }
                        WriteLn (Format (ABC_IO_COMPRESSION_TIME_TEXT, [Execution_Time]));
                        WriteLn (ABC_IO_EMPTY_TEXT);
                      End; { then }
                  End; { For }
              Finally
                ABC_MCP.Free;
              End; { Try }
            End { then }
              else
            Begin { else }
              WriteLn (ABC_IO_FILE_NOT_FOUND);
              Commands.Compress_F := FALSE;
              Commands.Expand_F   := FALSE;
            End; { else }
        Finally
          File_List.Free;
        End; { Try }
      End; { then }

    { EXPAND }
    If Commands.Expand_F = TRUE then
      Begin { then }
        { Expand all files }
        File_List := TStringList.Create;
        Try
          ABC_IO.Get_Filenames (Commands.Input_Filename, File_List);
          File_N  := File_List.Count;
          If Commands.Verbose_F = FALSE then
            Begin { then }
              Write (Format (ABC_IO_EXPAND_START_TEXT, [File_N]));
            End; { then }

          ABC_MCP := T_ABC_MCP.Create;
          Try
            For I := 0 to File_List.Count - 1 do
              Begin { For }
                { Get input filename and output filename }
                Input_Filename  := File_List.Strings [I];
                If ExtractFileExt (Input_Filename) <> '' then
                  Begin { then }
                    Output_Filename := Copy (Input_Filename, 1, Pos (ExtractFileExt (Input_Filename), Input_Filename)) + Commands.Output_Extension;
                  End { then }
                    else
                  Begin { else }
                    Output_Filename := Input_Filename + ABC_IO_DELIMITER + Commands.Output_Extension;
                  End; { else }
                If Input_Filename = Output_Filename then
                  Begin { then }
                    Output_Filename := Output_Filename + ABC_IO_DELIMITER + Commands.Output_Extension
                  End; { then }

                { Print command }
                S := ABC_IO_EXPAND_TEXT + Input_Filename + ABC_IO_BETWEEN_TEXT + Output_Filename;
                If Commands.Verbose_F = TRUE then
                  Begin { then }
                    WriteLn (S);
                  End { then }
                    else
                  Begin { else }
                    Write (ABC_A_IO_WAIT_TEXT [I mod ABC_IO_WAIT_TEXT_N]);
                  End; { else }

                { Expand }
                Starttime := DateTimeToTimeStamp (Now);
                ABC_MCP.Expand_File (Input_Filename, Output_Filename);
                Endtime := DateTimeToTimeStamp (Now);

                { Print stages }
                If ABC_MCP.Compress_Information.Fast_F = TRUE then
                  Begin { then }
                    S := ABC_IO_MTF_EXPAND_TEXT;
                  End { then }
                    else
                  Begin { else }
                    If ABC_MCP.Compress_Information.SIF_F = TRUE then
                      Begin { then }
                        S := ABC_IO_SIF_EXPAND_TEXT;
                      End { then }
                        else
                      Begin { else }
                        S := ABC_IO_AWFC_EXPAND_TEXT;
                      End; { else }
                  End; { else }
                If Commands.Verbose_F = TRUE then
                  Begin { then }
                    WriteLn (S);
                  End; { then }

                { Print input size and output size }
                Input_Size        := ABC_IO.Get_File_Size (Input_Filename);
                Total_Input_Size  := Total_Input_Size + Input_Size;
                Output_Size       := ABC_IO.Get_File_Size (Output_Filename);
                Total_Output_Size := Total_Output_Size + Output_Size;
                If Commands.Verbose_F = TRUE then
                  Begin { then }
                    WriteLn (Format (ABC_IO_INPUT_SIZE_TEXT, [Input_Size]));
                    WriteLn (Format (ABC_IO_OUTPUT_SIZE_TEXT, [Output_Size]));
                  End; { then }

                { Print compression rate }
                If Input_Size <> 0 then
                  Begin { then }
                    { Don't devide by zero }
                    Compression_Rate := Input_Size / Output_Size * ABC_BYTE_BIT_SIZE;
                    If Commands.Verbose_F = TRUE then
                      Begin { then }
                        WriteLn (Format (ABC_IO_COMPRESSION_RATIO_TEXT, [Compression_Rate]));
                      End; { then }
                    Average_Compression_Rate := Average_Compression_Rate + Compression_Rate;
                  End; { then }

                { Print expansion time }
                Execution_Time  := (TimeStampToMSecs (Endtime) - TimeStampToMSecs (Starttime)) / 1000;
                If Commands.Verbose_F = TRUE then
                  Begin { then }
                    WriteLn (Format (ABC_IO_EXPANSION_TIME_TEXT, [Execution_Time]));
                    WriteLn (Format (ABC_IO_CRC_VALID, []));
                    WriteLn (ABC_IO_EMPTY_TEXT);
                  End; { then }
              End; { For }
          Finally
            ABC_MCP.Free;
          End; { Try }
        Finally
          File_List.Free;
        End; { Try }
      End; { then }
  Finally
    ABC_IO.Free;
  End; { Try }

  If (Commands.Compress_F = TRUE) or (Commands.Expand_F = TRUE) then
    Begin { then }
      { Print statistic }
      Total_Endtime   := DateTimeToTimeStamp (Now);
      Execution_Time  := (TimeStampToMSecs (Total_Endtime) - TimeStampToMSecs (Total_Starttime)) / 1000;

      If Commands.Verbose_F = TRUE then
        Begin { then }
          For I := 0 to ABC_IO_STATISTIC_TEXT_N - 2 do
            Begin { For }
              WriteLn (ABC_A_IO_STATISTIC_TEXT [I]);
            End; { For }
        End { then }
          else
        Begin { else }
          Write (ABC_CR_CHR);
        End; { else }

      WriteLn (Format (ABC_IO_NUMBER_OF_FILES_TEXT, [File_N]));
      WriteLn (Format (ABC_IO_TOTAL_INPUT_SIZE_TEXT, [Total_Input_Size]));
      WriteLn (Format (ABC_IO_TOTAL_OUTPUT_SIZE_TEXT, [Total_Output_Size]));
      WriteLn (Format (ABC_IO_AVERAGE_COMPRESSION_RATE_TEXT, [Average_Compression_Rate / File_N]));
      If Total_Input_Size <> 0 then
        Begin { then }
          { Don't devide by zero }
          If Commands.Compress_F = TRUE then
            Begin { then }
              Compression_Rate := Total_Output_Size / Total_Input_Size * ABC_BYTE_BIT_SIZE;
            End { then }
              else
            Begin { else }
              Compression_Rate := Total_Input_Size / Total_Output_Size * ABC_BYTE_BIT_SIZE;
            End; { else }
          WriteLn (Format (ABC_IO_TOTAL_COMPRESSION_RATE_TEXT, [Compression_Rate]));
        End; { then }
      WriteLn (Format (ABC_IO_TOTAL_TIME_TEXT, [Execution_Time]));
      WriteLn (ABC_A_IO_STATISTIC_TEXT [ABC_IO_STATISTIC_TEXT_N - 1]);
    End; { then }

  {$IFNDEF NO_WAIT}
  { Wait for RETURN }
  ReadLn;
  {$ENDIF}
End.
