{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_IO_Unit                                                                     ##}
{##                                                                                 ##}
{## I/O procedures and a simple command line parser                                 ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_IO_Unit;


Interface


Uses
  Classes,
  SysUtils,
  ABC_Global_Unit;


Const
  ABC_IO_PARAMETER_COMMAND_MAX = 6;


Type
  T_ABC_IO = Class
  Public
    Procedure Parse_Command_Line  (Var Commands : T_R_ABC_Commands);
    Procedure Get_Filenames (Const File_Specifier : String; Var File_List : TStringList);
    Function  Get_File_Size (Const File_Name : String) : Integer;
  End; { T_ABC_IO }


Implementation


{-------------------------------------------------------------------------------------}
Procedure T_ABC_IO.Parse_Command_Line (Var Commands : T_R_ABC_Commands);
{ Command line parser                                                                 }
{-------------------------------------------------------------------------------------}
  Var
    Parameter_Command : String;
    I                 : Integer;
    J                 : Integer;

  Begin { T_ABC_IO.Parse_Command_Line }
    { Initialize commands }
    With Commands do
      Begin { With }
        Help_F            := FALSE;
        Compress_F        := FALSE;
        Expand_F          := FALSE;
        Verbose_F         := FALSE;
        Fast_F            := FALSE;
        Input_Filename    := '';
        Output_Extension  := '';
      End; { With }

    { Parse all parameters }
    For I := 0 to ABC_IO_PARAMETER_COMMAND_MAX - 1 do
      Begin { For }
        Parameter_Command := AnsiLowerCase (ParamStr (I + 1));
        If Parameter_Command <> '' then
          Begin { then }
            If Copy (Parameter_Command, 1, 1) = ABC_IO_OPTION_START then
              Begin { then }
                For J := 0 to ABC_IO_PARAMETER_COMMAND_MAX - 1 do
                  Begin { For }
                    { Parse help }
                    If Copy (Parameter_Command, 2, 1) = ABC_IO_OPTION_HELP then
                      Begin { then }
                        Commands.Help_F     := TRUE;
                        Parameter_Command   := ABC_IO_OPTION_START + Copy (Parameter_Command, 3, ABC_IO_PARAMETER_COMMAND_MAX);
                      End; { then }

                    { Parse compress }
                    If Copy (Parameter_Command, 2, 1) = ABC_IO_OPTION_COMPRESS then
                      Begin { then }
                        Commands.Compress_F       := TRUE;
                        Parameter_Command         := ABC_IO_OPTION_START + Copy (Parameter_Command, 3, ABC_IO_PARAMETER_COMMAND_MAX);
                        Commands.Output_Extension := ABC_IO_DEFAULT_COMPRESS_EXTENSION;
                      End; { then }

                    { Parse expand }
                    If Copy (Parameter_Command, 2, 1) = ABC_IO_OPTION_EXPAND then
                      Begin { then }
                        Commands.Expand_F         := TRUE;
                        Parameter_Command         := ABC_IO_OPTION_START + Copy (Parameter_Command, 3, ABC_IO_PARAMETER_COMMAND_MAX);
                        Commands.Output_Extension := ABC_IO_DEFAULT_EXPAND_EXTENSION;
                      End; { then }

                    { Parse verbose }
                    If Copy (Parameter_Command, 2, 1) = ABC_IO_OPTION_VERBOSE then
                      Begin { then }
                        Commands.Verbose_F  := TRUE;
                        Parameter_Command   := ABC_IO_OPTION_START + Copy (Parameter_Command, 3, ABC_IO_PARAMETER_COMMAND_MAX);
                      End; { then }

                    { Parse fast }
                    If Copy (Parameter_Command, 2, 1) = ABC_IO_OPTION_FAST then
                      Begin { then }
                        Commands.Fast_F     := TRUE;
                        Parameter_Command   := ABC_IO_OPTION_START + Copy (Parameter_Command, 3, ABC_IO_PARAMETER_COMMAND_MAX);
                      End; { then }
                  End; { For }
              End { then }
                else
              Begin { else }
                { Parse input filename and extension }
                If Commands.Input_Filename = '' then
                  Begin { then }
                    Commands.Input_Filename   := Parameter_Command;
                  End { then }
                    else
                  Begin { else }
                    Commands.Output_Extension := Parameter_Command;
                  End; { else }
              End; { else }
          End; { then }
      End; { For }

    { Check for no or wrong option }
    If (Commands.Help_F = FALSE) and (Commands.Compress_F = FALSE) and (Commands.Expand_F = FALSE) then
      Begin { then }
        Commands.Help_F := TRUE;
      End; { then }
  End; { T_ABC_IO.Parse_Command_Line }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_IO.Get_Filenames (Const File_Specifier : String; Var File_List : TStringList);
{ Returns all filenames of the specified files                                        }
{-------------------------------------------------------------------------------------}
  Var
    I             : Integer;
    Search_Record : TSearchRec;

  Begin { T_ABC_IO.Get_Filenames }
    { Get first filename }
    I := FindFirst (File_Specifier, faAnyFile - faVolumeID - faDirectory, Search_Record);

    { Fill File_List }
    While I = 0 do
      Begin { While }
        File_List.Add (ExtractFilePath (File_Specifier) + Search_Record.Name);

        { Get next filename }
        I := FindNext (Search_Record);
      End; { While }
    FindClose (Search_Record);
  End; { T_ABC_IO.Get_Filenames }


{-------------------------------------------------------------------------------------}
Function T_ABC_IO.Get_File_Size (Const File_Name : String) : Integer;
{ Returns the size of a file                                                          }
{-------------------------------------------------------------------------------------}
  Var
    Search_Record : TSearchRec;

  Begin { T_ABC_IO.Get_File_Size }
    If FindFirst (File_Name, faAnyFile, Search_Record) = 0 then
      Begin { then }
        Result := Search_Record.Size;
      End { then }
        else
      Begin { else }
        Result := -1;
      End; { else }
    FindClose (Search_Record);
  End; { T_ABC_IO.Get_File_Size }

End.

