{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_MTF_Unit                                                                    ##}
{##                                                                                 ##}
{## Move-To-Front transformation                                                    ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_MTF_Unit;


Interface


Uses
  Classes, ABC_Global_Unit;


Const
  ABC_MTF_BUFFER_SIZE  = 3; { Size of symbol buffer and number of tested symbols }
  ABC_MTF_NEW_POSITION = 1; { Position of symbols, which are not referenced before }


Type
  T_ABC_MTF = Class
  Protected
    M_A_Symbol_To_Index   : Array [0 .. ABC_BYTE_MAXIMUM_SYMBOL] of Longword;
    M_A_Symbol_Buffer     : Array [0 .. ABC_BYTE_MAXIMUM_SYMBOL] of Longword;
    M_A_Last_Mapping      : Array [0 .. ABC_MTF_BUFFER_SIZE - 1] of Longword;
    Procedure   Init;
    Procedure   Move (Symbol : Longword);
  Public
    Function    Encode_Symbol (Symbol : Longword) : Longword;
    Function    Decode_Symbol (Position : Longword) : Longword;
    Procedure   Encode (Var Compress_Information : T_R_ABC_Compress_Information);
    Procedure   Decode (Var Compress_Information : T_R_ABC_Compress_Information);
  End; { T_ABC_MTF }


Implementation


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MTF.Init;
{ Initializing of array and variables                                                 }
{-------------------------------------------------------------------------------------}
  Var
    I : Longword;

  Begin { T_ABC_MTF.Init }
    For I := 0 to ABC_BYTE_MAXIMUM_SYMBOL do
      Begin { For }
        M_A_Symbol_Buffer [I]   := I;
        M_A_Symbol_To_Index [I] := I;
      End; { For }
    FillChar (M_A_Last_Mapping [0], SizeOf (M_A_Last_Mapping), $00);
  End; { T_ABC_MTF.Init }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MTF.Move (Symbol : Longword);
{ Move symbol by updating table                                                       }
{-------------------------------------------------------------------------------------}
  Var
    I             : Longword;
    Current_Index : Longword;
    {$IFDEF MTF11}
    B             : Boolean;
    {$ENDIF}

  Begin { T_ABC_MTF.Move }
    Current_Index := M_A_Symbol_To_Index [Symbol];
    If Current_Index <> 0 then
      Begin { then }
        {$IFDEF MTF0}
        For I := Current_Index downto 1 do
          Begin { For }
            M_A_Symbol_Buffer [I]                       := M_A_Symbol_Buffer [I - 1];
            M_A_Symbol_To_Index [M_A_Symbol_Buffer [I]] := I;
          End; { For }
        M_A_Symbol_Buffer [0]         := Symbol;
        M_A_Symbol_To_Index [Symbol]  := 0;
        {$ENDIF}
        {$IFDEF MTF1}
        If Current_Index = ABC_MTF_NEW_POSITION then
          Begin { then }
            If M_A_Last_Mapping [0] <> $00 then
              Begin { then }
                M_A_Symbol_Buffer [ABC_MTF_NEW_POSITION]                       := M_A_Symbol_Buffer [0];
                M_A_Symbol_To_Index [M_A_Symbol_Buffer [ABC_MTF_NEW_POSITION]] := ABC_MTF_NEW_POSITION;
                M_A_Symbol_Buffer [0]                                          := Symbol;
                M_A_Symbol_To_Index [Symbol]                                   := 0;
              End; { then }
          End { then }
            else
          Begin { else }
            For I := Current_Index downto 2 do
              Begin { For }
                M_A_Symbol_Buffer [I]                       := M_A_Symbol_Buffer [I - 1];
                M_A_Symbol_To_Index [M_A_Symbol_Buffer [I]] := I;
              End; { For }
            M_A_Symbol_Buffer [1]         := Symbol;
            M_A_Symbol_To_Index [Symbol]  := 1;
          End; { else }
        {$ENDIF}
        {$IFDEF MTF11}
        If Current_Index = ABC_MTF_NEW_POSITION then
          Begin { then }
            B := TRUE;
            For I := 0 to ABC_MTF_BUFFER_SIZE - 1 do
              Begin { For }
                If M_A_Last_Mapping [I] < ABC_MTF_NEW_POSITION then
                  Begin { then }
                    B := FALSE;
                    Break;
                  End; { then }
              End; { For }
            If B = TRUE then
              Begin { then }
                M_A_Symbol_Buffer [ABC_MTF_NEW_POSITION]                       := M_A_Symbol_Buffer [0];
                M_A_Symbol_To_Index [M_A_Symbol_Buffer [ABC_MTF_NEW_POSITION]] := ABC_MTF_NEW_POSITION;
                M_A_Symbol_Buffer [0]                                          := Symbol;
                M_A_Symbol_To_Index [Symbol]                                   := 0;
              End; { then }
          End { then }
            else
          Begin { else }
            For I := Current_Index downto ABC_MTF_NEW_POSITION + 1 do
              Begin { For }
                M_A_Symbol_Buffer [I]                       := M_A_Symbol_Buffer [I - 1];
                M_A_Symbol_To_Index [M_A_Symbol_Buffer [I]] := I;
              End; { For }
            M_A_Symbol_Buffer [ABC_MTF_NEW_POSITION] := Symbol;
            M_A_Symbol_To_Index [Symbol]             := ABC_MTF_NEW_POSITION;
          End; { else }
        {$ENDIF}
      End; { then }

    {$IFNDEF MTF0}
    { Save last symbols }
    For I := ABC_MTF_BUFFER_SIZE - 1 downto 1 do
      Begin { For }
        M_A_Last_Mapping [I] := M_A_Last_Mapping [I - 1];
      End; { For }
    M_A_Last_Mapping [0] := Current_Index;
    {$ENDIF}
  End; { T_ABC_MTF.Move }


{-------------------------------------------------------------------------------------}
Function T_ABC_MTF.Encode_Symbol (Symbol : Longword) : Longword;
{ Codes a symbol and updates the table                                                }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_MTF.Encode_Symbol }
    Result := M_A_Symbol_To_Index [Symbol];
    Move (Symbol);
  End; { T_ABC_MTF.Encode_Symbol }


{-------------------------------------------------------------------------------------}
Function T_ABC_MTF.Decode_Symbol (Position : Longword) : Longword;
{ Decodes a symbol and updates the table                                              }
{-------------------------------------------------------------------------------------}
  Begin { T_ABC_MTF.Decode_Symbol }
    Result := M_A_Symbol_Buffer [Position];
    Move (Result);
  End; { T_ABC_MTF.Decode_Symbol }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MTF.Encode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Codes a block                                                                       }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_MTF.Encode }
    Init; { Initializing data }
    For I := 0 to Compress_Information.P_Input_Record^.Size - 1 do
      Begin { For }
        Compress_Information.P_Output_Record^.P_Buffer^ [I] := Encode_Symbol (Compress_Information.P_Input_Record^.P_Buffer^ [I]);
      End; { For }

    { Set length }
    Compress_Information.P_Output_Record^.Size := Compress_Information.P_Input_Record^.Size;
  End; { T_ABC_MTF.Encode }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_MTF.Decode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Decodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    I : Integer;

  Begin { T_ABC_MTF.Decode }
    Init; { Initializing data }
    For I := 0 to Compress_Information.P_Input_Record^.Size - 1 do
      Begin { For }
        Compress_Information.P_Output_Record^.P_Buffer^ [I] := Decode_Symbol (Compress_Information.P_Input_Record^.P_Buffer^ [I]);
      End; { For }

    { Set length }
    Compress_Information.P_Output_Record^.Size := Compress_Information.P_Input_Record^.Size;
  End; { T_ABC_MTF.Decode }


End.

