{#####################################################################################}
{##                                                                                 ##}
{## ABC - Advanced Blocksorting Compressor                                          ##}
{##                                                                                 ##}
{## ABC_RLE_BIT_Unit                                                                ##}
{##                                                                                 ##}
{## Run-Length-Encoding with binary coding of the runlength                         ##}
{##                                                                                 ##}
{## Copyright (C) 2002-2003 Jrgen Abel                                             ##}
{##                                                                                 ##}
{#####################################################################################}


{$I ABC_Include.pas}


Unit ABC_RLE_BIT_Unit;


Interface


Uses
  ABC_Global_Unit;


Type
  T_ABC_RLE_BIT = Class
  Protected
    M_Threshold : Integer;
  Public
    Procedure   Encode (Var Compress_Information : T_R_ABC_Compress_Information);
    Procedure   Decode (Var Compress_Information : T_R_ABC_Compress_Information);
    Property    Threshold : Integer Read M_Threshold Write M_Threshold;
  End; { T_ABC_RLE_BIT }


Implementation


Uses
  ABC_MCP_Unit;


{-------------------------------------------------------------------------------------}
Procedure T_ABC_RLE_BIT.Encode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Encodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    Input_Index     : Integer;
    Output_Index    : Integer;
    Current_Symbol  : Byte;
    Run_Symbol      : Byte;
    Run_Start_Index : Integer;
    Run_Length      : Integer;

  Begin { T_ABC_RLE_BIT.Encode }
    If Compress_Information.P_Input_Record^.Size <= 0 then
      Begin { then }
        { Switch buffer }
        ABC_MCP.Switch_Block (Compress_Information);
        Exit;
      End; { then }

    Input_Index     := 0;
    Output_Index    := 0;

    { Copy first symbol }
    Run_Symbol                                                      := Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];
    Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Run_Symbol;
    Compress_Information.P_Temp_A_0^ [Output_Index]                 := 0;
    Run_Start_Index                                                 := Input_Index;
    Inc (Output_Index);
    Inc (Input_Index);

    { Handle symbols }
    While Input_Index <= Compress_Information.P_Input_Record^.Size - 1 do
      Begin { While }
        Current_Symbol := Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];
        If Current_Symbol = Run_Symbol then
          Begin { then }
            Run_Length := (Input_Index + 1) - Run_Start_Index;
            If Run_Length < M_Threshold then
              Begin { then }
                { Output symbol }
                Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Run_Symbol;
                Compress_Information.P_Temp_A_0^ [Output_Index]                 := 0;
                Inc (Output_Index);
              End; { then }
          End { then }
            else
          Begin { else }
            Run_Length := Input_Index - Run_Start_Index;
            If Run_Length >= M_Threshold then
              Begin { then }
                { Output run length }
                Compress_Information.P_Temp_A_0^ [Output_Index - 1] := Run_Length - M_Threshold + 1;
              End; { then }

            { Output symbol }
            Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Current_Symbol;
            Compress_Information.P_Temp_A_0^ [Output_Index]                 := 0;
            Inc (Output_Index);
            Run_Symbol                                                      := Current_Symbol;
            Run_Start_Index                                                 := Input_Index;
          End; { else }
        Inc (Input_Index);
      End; { While }

    { Encode last run }
    Run_Length := Input_Index - Run_Start_Index;
    If Run_Length >= M_Threshold then
      Begin { then }
        { Output run length }
        Compress_Information.P_Temp_A_0^ [Output_Index - 1] := Run_Length - M_Threshold + 1;
      End; { then }

    { Set length }
    Compress_Information.P_Output_Record^.Size := Output_Index;
  End; { T_ABC_RLE_BIT.Encode }


{-------------------------------------------------------------------------------------}
Procedure T_ABC_RLE_BIT.Decode (Var Compress_Information : T_R_ABC_Compress_Information);
{ Decodes a block                                                                     }
{-------------------------------------------------------------------------------------}
  Var
    Input_Index   : Integer;
    Output_Index  : Integer;
    Symbol        : Integer;
    Run_Length    : Integer;
    I             : Integer;

  Begin { T_ABC_RLE_BIT.Decode }
    If Compress_Information.P_Input_Record^.Size <= 0 then
      Begin { then }
        { Switch buffer }
        ABC_MCP.Switch_Block (Compress_Information);
        Exit;
      End; { then }

    { Handle all symbols }
    Input_Index  := 0;
    Output_Index := 0;
    While Input_Index <= Compress_Information.P_Input_Record^.Size - 1 do
      Begin { While }
        { Copy symbol }
        Symbol                                                          := Compress_Information.P_Input_Record^.P_Buffer^ [Input_Index];
        Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Symbol;
        Inc (Output_Index);
        Inc (Input_Index);

        { Get run length }
        Run_Length := Compress_Information.P_Temp_A_0^ [Input_Index - 1];
        For I := 0 to Run_Length - 1 do
          Begin { For }
            Compress_Information.P_Output_Record^.P_Buffer^ [Output_Index]  := Symbol;
            Inc (Output_Index);
          End; { For }
      End; { While }

    { Set length }
    Compress_Information.P_Output_Record^.Size := Output_Index;
  End; { T_ABC_RLE_BIT.Decode }


End.

