
{*******************************************************}
{                                                       }
{       MessageInterceptor component                    }
{                                                       }
{       By Pablo Reyes preyes@excite.com                }
{       Freeware. Use it. Enjoy it!                     }
{                                                       }
{  Instructions:                                        }
{    - Set Control property                             }
{    - Set Active to True                               }
{    - When the related control receive a message       }
{      BeforeMessage and AfterMessage events occurs.    }
{    - In BeforeMessage event set Continue to False     }
{      to avoid passing message to related control.     }
{                                                       }
{*******************************************************}

unit MessageInterceptor;

interface

uses
	Classes, Controls, Messages, SysUtils;

type
	TBeforeMessageEvent = procedure(Sender: TObject; var Msg: TMessage;
		var Continue: boolean) of object;
	TAfterMessageEvent = procedure(Sender: TObject; var Msg: TMessage) of object;

	TMessageInterceptor = class(TComponent)
	private
		FActive: boolean;
		FControl: TControl;
		FDefaultWindowProc: TWndMethod; // WindowProc of the associated control
		FBeforeMessage: TBeforeMessageEvent;
		FAfterMessage: TAfterMessageEvent;
		procedure SetActive(const Value: boolean);
		procedure SetControl(const Value: TControl);
		function GetControl: TControl;
		function GetActive: boolean;
	protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
		procedure _WindowProc(var Msg: TMessage); virtual; // Component WindowProc
		procedure AssignWindowProc;
	public
		function MessageText(Msg: cardinal): string;
		property DefaultWindowProc: TWndMethod read FDefaultWindowProc
			write FDefaultWindowProc;
	published
		property Active: boolean read GetActive write SetActive;
		property Control: TControl read GetControl write SetControl;
		property BeforeMessage: TBeforeMessageEvent read FBeforeMessage
			write FBeforeMessage;
		property AfterMessage: TAfterMessageEvent read FAfterMessage
			write FAfterMessage;
	end;

	procedure Register;

implementation

uses
	Dialogs;

procedure Register;
begin
	RegisterComponents('Samples', [TMessageInterceptor]);
end;

{ TDropFiles }

procedure TMessageInterceptor.AssignWindowProc;
begin
	FDefaultWindowProc := FControl.WindowProc;
	FControl.WindowProc := _WindowProc;
end;

function TMessageInterceptor.GetActive: boolean;
begin
	Result := FActive;
end;

function TMessageInterceptor.GetControl: TControl;
begin
	Result := FControl;
end;

function TMessageInterceptor.MessageText(Msg: cardinal): string;
begin
	case Msg of
		$0000: Result := 'WM_NULL';
		$0001: Result := 'WM_CREATE';
		$0002: Result := 'WM_DESTROY';
		$0003: Result := 'WM_MOVE';
		$0005: Result := 'WM_SIZE';
		$0006: Result := 'WM_ACTIVATE';
		$0007: Result := 'WM_SETFOCUS';
		$0008: Result := 'WM_KILLFOCUS';
		$000A: Result := 'WM_ENABLE';
		$000B: Result := 'WM_SETREDRAW';
		$000C: Result := 'WM_SETTEXT';
		$000D: Result := 'WM_GETTEXT';
		$000E: Result := 'WM_GETTEXTLENGTH';
		$000F: Result := 'WM_PAINT';
		$0010: Result := 'WM_CLOSE';
		$0011: Result := 'WM_QUERYENDSESSION';
		$0012: Result := 'WM_QUIT';
		$0013: Result := 'WM_QUERYOPEN';
		$0014: Result := 'WM_ERASEBKGND';
		$0015: Result := 'WM_SYSCOLORCHANGE';
		$0016: Result := 'WM_ENDSESSION';
		$0017: Result := 'WM_SYSTEMERROR';
		$0018: Result := 'WM_SHOWWINDOW';
		$0019: Result := 'WM_CTLCOLOR';
		$001A: Result := 'WM_WININICHANGE / WM_WININICHANGE';
		$001B: Result := 'WM_DEVMODECHANGE';
		$001C: Result := 'WM_ACTIVATEAPP';
		$001D: Result := 'WM_FONTCHANGE';
		$001E: Result := 'WM_TIMECHANGE';
		$001F: Result := 'WM_CANCELMODE';
		$0020: Result := 'WM_SETCURSOR';
		$0021: Result := 'WM_MOUSEACTIVATE';
		$0022: Result := 'WM_CHILDACTIVATE';
		$0023: Result := 'WM_QUEUESYNC';
		$0024: Result := 'WM_GETMINMAXINFO';
		$0026: Result := 'WM_PAINTICON';
		$0027: Result := 'WM_ICONERASEBKGND';
		$0028: Result := 'WM_NEXTDLGCTL';
		$002A: Result := 'WM_SPOOLERSTATUS';
		$002B: Result := 'WM_DRAWITEM';
		$002C: Result := 'WM_MEASUREITEM';
		$002D: Result := 'WM_DELETEITEM';
		$002E: Result := 'WM_VKEYTOITEM';
		$002F: Result := 'WM_CHARTOITEM';
		$0030: Result := 'WM_SETFONT';
		$0031: Result := 'WM_GETFONT';
		$0032: Result := 'WM_SETHOTKEY';
		$0033: Result := 'WM_GETHOTKEY';
		$0037: Result := 'WM_QUERYDRAGICON';
		$0039: Result := 'WM_COMPAREITEM';
		$003D: Result := 'WM_GETOBJECT';
		$0041: Result := 'WM_COMPACTING';
		$0046: Result := 'WM_WINDOWPOSCHANGING';
		$0047: Result := 'WM_WINDOWPOSCHANGED';
		$0048: Result := 'WM_POWER';
		$004A: Result := 'WM_COPYDATA';
		$004B: Result := 'WM_CANCELJOURNAL';
		$004E: Result := 'WM_NOTIFY';
		$0050: Result := 'WM_INPUTLANGCHANGEREQUEST';
		$0051: Result := 'WM_INPUTLANGCHANGE';
		$0052: Result := 'WM_TCARD';
		$0053: Result := 'WM_HELP';
		$0054: Result := 'WM_USERCHANGED';
		$0055: Result := 'WM_NOTIFYFORMAT';
		$007B: Result := 'WM_CONTEXTMENU';
		$007C: Result := 'WM_STYLECHANGING';
		$007D: Result := 'WM_STYLECHANGED';
		$007E: Result := 'WM_DISPLAYCHANGE';
		$007F: Result := 'WM_GETICON';
		$0080: Result := 'WM_SETICON';
		$0081: Result := 'WM_NCCREATE';
		$0082: Result := 'WM_NCDESTROY';
		$0083: Result := 'WM_NCCALCSIZE';
		$0084: Result := 'WM_NCHITTEST';
		$0085: Result := 'WM_NCPAINT';
		$0086: Result := 'WM_NCACTIVATE';
		$0087: Result := 'WM_GETDLGCODE';
		$00A0: Result := 'WM_NCMOUSEMOVE';
		$00A1: Result := 'WM_NCLBUTTONDOWN';
		$00A2: Result := 'WM_NCLBUTTONUP';
		$00A3: Result := 'WM_NCLBUTTONDBLCLK';
		$00A4: Result := 'WM_NCRBUTTONDOWN';
		$00A5: Result := 'WM_NCRBUTTONUP';
		$00A6: Result := 'WM_NCRBUTTONDBLCLK';
		$00A7: Result := 'WM_NCMBUTTONDOWN';
		$00A8: Result := 'WM_NCMBUTTONUP';
		$00A9: Result := 'WM_NCMBUTTONDBLCLK / WM_KEYFIRST';
		$0100: Result := 'WM_KEYDOWN';
		$0101: Result := 'WM_KEYUP';
		$0102: Result := 'WM_CHAR';
		$0103: Result := 'WM_DEADCHAR';
		$0104: Result := 'WM_SYSKEYDOWN';
		$0105: Result := 'WM_SYSKEYUP';
		$0106: Result := 'WM_SYSCHAR';
		$0107: Result := 'WM_SYSDEADCHAR';
		$0108: Result := 'WM_KEYLAST';
		$0110: Result := 'WM_INITDIALOG';
		$0111: Result := 'WM_COMMAND';
		$0112: Result := 'WM_SYSCOMMAND';
		$0113: Result := 'WM_TIMER';
		$0114: Result := 'WM_HSCROLL';
		$0115: Result := 'WM_VSCROLL';
		$0116: Result := 'WM_INITMENU';
		$0117: Result := 'WM_INITMENUPOPUP';
		$011F: Result := 'WM_MENUSELECT';
		$0120: Result := 'WM_MENUCHAR';
		$0121: Result := 'WM_ENTERIDLE';
		$0122: Result := 'WM_MENURBUTTONUP';
		$0123: Result := 'WM_MENUDRAG';
		$0124: Result := 'WM_MENUGETOBJECT';
		$0125: Result := 'WM_UNINITMENUPOPUP';
		$0126: Result := 'WM_MENUCOMMAND';
		$0132: Result := 'WM_CTLCOLORMSGBOX';
		$0133: Result := 'WM_CTLCOLOREDIT';
		$0134: Result := 'WM_CTLCOLORLISTBOX';
		$0135: Result := 'WM_CTLCOLORBTN';
		$0136: Result := 'WM_CTLCOLORDLG';
		$0137: Result := 'WM_CTLCOLORSCROLLBAR';
		$0138: Result := 'WM_CTLCOLORSTATIC';
		$0200: Result := 'WM_MOUSEFIRST / WM_MOUSEMOVE';
		$0201: Result := 'WM_LBUTTONDOWN';
		$0202: Result := 'WM_LBUTTONUP';
		$0203: Result := 'WM_LBUTTONDBLCLK';
		$0204: Result := 'WM_RBUTTONDOWN';
		$0205: Result := 'WM_RBUTTONUP';
		$0206: Result := 'WM_RBUTTONDBLCLK';
		$0207: Result := 'WM_MBUTTONDOWN';
		$0208: Result := 'WM_MBUTTONUP';
		$0209: Result := 'WM_MBUTTONDBLCLK';
		$020A: Result := 'WM_MOUSEWHEEL / WM_MOUSELAST';
		$0210: Result := 'WM_PARENTNOTIFY';
		$0211: Result := 'WM_ENTERMENULOOP';
		$0212: Result := 'WM_EXITMENULOOP';
		$0213: Result := 'WM_NEXTMENU';
		532: Result := 'WM_SIZING';
		533: Result := 'WM_CAPTURECHANGED';
		534: Result := 'WM_MOVING';
		536: Result := 'WM_POWERBROADCAST';
		537: Result := 'WM_DEVICECHANGE';
		$010D: Result := 'WM_IME_STARTCOMPOSITION';
		$010E: Result := 'WM_IME_ENDCOMPOSITION';
		$010F: Result := 'WM_IME_COMPOSITION / WM_IME_KEYLAST';
		$0281: Result := 'WM_IME_SETCONTEXT';
		$0282: Result := 'WM_IME_NOTIFY';
		$0283: Result := 'WM_IME_CONTROL';
		$0284: Result := 'WM_IME_COMPOSITIONFULL';
		$0285: Result := 'WM_IME_SELECT';
		$0286: Result := 'WM_IME_CHAR';
		$0288: Result := 'WM_IME_REQUEST';
		$0290: Result := 'WM_IME_KEYDOWN';
		$0291: Result := 'WM_IME_KEYUP';
		$0220: Result := 'WM_MDICREATE';
		$0221: Result := 'WM_MDIDESTROY';
		$0222: Result := 'WM_MDIACTIVATE';
		$0223: Result := 'WM_MDIRESTORE';
		$0224: Result := 'WM_MDINEXT';
		$0225: Result := 'WM_MDIMAXIMIZE';
		$0226: Result := 'WM_MDITILE';
		$0227: Result := 'WM_MDICASCADE';
		$0228: Result := 'WM_MDIICONARRANGE';
		$0229: Result := 'WM_MDIGETACTIVE';
		$0230: Result := 'WM_MDISETMENU';
		$0231: Result := 'WM_ENTERSIZEMOVE';
		$0232: Result := 'WM_EXITSIZEMOVE';
		$0233: Result := 'WM_DROPFILES';
		$0234: Result := 'WM_MDIREFRESHMENU';
		$02A1: Result := 'WM_MOUSEHOVER';
		$02A3: Result := 'WM_MOUSELEAVE';
		$0300: Result := 'WM_CUT';
		$0301: Result := 'WM_COPY';
		$0302: Result := 'WM_PASTE';
		$0303: Result := 'WM_CLEAR';
		$0304: Result := 'WM_UNDO';
		$0305: Result := 'WM_RENDERFORMAT';
		$0306: Result := 'WM_RENDERALLFORMATS';
		$0307: Result := 'WM_DESTROYCLIPBOARD';
		$0308: Result := 'WM_DRAWCLIPBOARD';
		$0309: Result := 'WM_PAINTCLIPBOARD';
		$030A: Result := 'WM_VSCROLLCLIPBOARD';
		$030B: Result := 'WM_SIZECLIPBOARD';
		$030C: Result := 'WM_ASKCBFORMATNAME';
		$030D: Result := 'WM_CHANGECBCHAIN';
		$030E: Result := 'WM_HSCROLLCLIPBOARD';
		$030F: Result := 'WM_QUERYNEWPALETTE';
		$0310: Result := 'WM_PALETTEISCHANGING';
		$0311: Result := 'WM_PALETTECHANGED';
		$0312: Result := 'WM_HOTKEY';
		791: Result := 'WM_PRINT';
		792: Result := 'WM_PRINTCLIENT';
		856: Result := 'WM_HANDHELDFIRST';
		863: Result := 'WM_HANDHELDLAST';
		$0380: Result := 'WM_PENWINFIRST';
		$038F: Result := 'WM_PENWINLAST';
		$0390: Result := 'WM_COALESCE_FIRST';
		$039F: Result := 'WM_COALESCE_LAST';
		$03E0: Result := 'WM_DDE_FIRST / WM_DDE_INITIATE';
		WM_DDE_FIRST + 1: Result := 'WM_DDE_TERMINATE';
		WM_DDE_FIRST + 2: Result := 'WM_DDE_ADVISE';
		WM_DDE_FIRST + 3: Result := 'WM_DDE_UNADVISE';
		WM_DDE_FIRST + 4: Result := 'WM_DDE_ACK';
		WM_DDE_FIRST + 5: Result := 'WM_DDE_DATA';
		WM_DDE_FIRST + 6: Result := 'WM_DDE_REQUEST';
		WM_DDE_FIRST + 7: Result := 'WM_DDE_POKE';
		WM_DDE_FIRST + 8: Result := 'WM_DDE_EXECUTE / WM_DDE_LAST';

		$0180: Result := 'LB_ADDSTRING';
		$0181: Result := 'LB_INSERTSTRING';
		$0182: Result := 'LB_DELETESTRING';
		$0183: Result := 'LB_SELITEMRANGEEX';
		$0184: Result := 'LB_RESETCONTENT';
		$0185: Result := 'LB_SETSEL';
		$0186: Result := 'LB_SETCURSEL';
		$0187: Result := 'LB_GETSEL';
		$0188: Result := 'LB_GETCURSEL';
		$0189: Result := 'LB_GETTEXT';
		$018A: Result := 'LB_GETTEXTLEN';
		$018B: Result := 'LB_GETCOUNT';
		$018C: Result := 'LB_SELECTSTRING';
		$018D: Result := 'LB_DIR';
		$018E: Result := 'LB_GETTOPINDEX';
		$018F: Result := 'LB_FINDSTRING';
		$0190: Result := 'LB_GETSELCOUNT';
		$0191: Result := 'LB_GETSELITEMS';
		$0192: Result := 'LB_SETTABSTOPS';
		$0193: Result := 'LB_GETHORIZONTALEXTENT';
		$0194: Result := 'LB_SETHORIZONTALEXTENT';
		$0195: Result := 'LB_SETCOLUMNWIDTH';
		$0196: Result := 'LB_ADDFILE';
		$0197: Result := 'LB_SETTOPINDEX';
		$0198: Result := 'LB_GETITEMRECT';
		$0199: Result := 'LB_GETITEMDATA';
		$019A: Result := 'LB_SETITEMDATA';
		$019B: Result := 'LB_SELITEMRANGE';
		$019C: Result := 'LB_SETANCHORINDEX';
		$019D: Result := 'LB_GETANCHORINDEX';
		$019E: Result := 'LB_SETCARETINDEX';
		$019F: Result := 'LB_GETCARETINDEX';
		$01A0: Result := 'LB_SETITEMHEIGHT';
		$01A1: Result := 'LB_GETITEMHEIGHT';
		$01A2: Result := 'LB_FINDSTRINGEXACT';
		$01A5: Result := 'LB_SETLOCALE';
		$01A6: Result := 'LB_GETLOCALE';
		$01A7: Result := 'LB_SETCOUNT';
		$01A8: Result := 'LB_INITSTORAGE';
		$01A9: Result := 'LB_ITEMFROMPOINT';
		432: Result := 'LB_MSGMAX';

		$0140: Result := 'CB_GETEDITSEL';
		$0141: Result := 'CB_LIMITTEXT';
		$0142: Result := 'CB_SETEDITSEL';
		$0143: Result := 'CB_ADDSTRING';
		$0144: Result := 'CB_DELETESTRING';
		$0145: Result := 'CB_DIR';
		$0146: Result := 'CB_GETCOUNT';
		$0147: Result := 'CB_GETCURSEL';
		$0148: Result := 'CB_GETLBTEXT';
		$0149: Result := 'CB_GETLBTEXTLEN';
		$014A: Result := 'CB_INSERTSTRING';
		$014B: Result := 'CB_RESETCONTENT';
		$014C: Result := 'CB_FINDSTRING';
		$014D: Result := 'CB_SELECTSTRING';
		$014E: Result := 'CB_SETCURSEL';
		$014F: Result := 'CB_SHOWDROPDOWN';
		$0150: Result := 'CB_GETITEMDATA';
		$0151: Result := 'CB_SETITEMDATA';
		$0152: Result := 'CB_GETDROPPEDCONTROLRECT';
		$0153: Result := 'CB_SETITEMHEIGHT';
		$0154: Result := 'CB_GETITEMHEIGHT';
		$0155: Result := 'CB_SETEXTENDEDUI';
		$0156: Result := 'CB_GETEXTENDEDUI';
		$0157: Result := 'CB_GETDROPPEDSTATE';
		$0158: Result := 'CB_FINDSTRINGEXACT';
		345: Result := 'CB_SETLOCALE';
		346: Result := 'CB_GETLOCALE';
		347: Result := 'CB_GETTOPINDEX';
		348: Result := 'CB_SETTOPINDEX';
		349: Result := 'CB_GETHORIZONTALEXTENT';
		350: Result := 'CB_SETHORIZONTALEXTENT';
		351: Result := 'CB_GETDROPPEDWIDTH';
		352: Result := 'CB_SETDROPPEDWIDTH';
		353: Result := 'CB_INITSTORAGE';
		354: Result := 'CB_MSGMAX';

		$0500: Result := 'EN_ERRSPACE';
		$0501: Result := 'EN_MAXTEXT';
		$0601: Result := 'EN_HSCROLL';
		$0602: Result := 'EN_VSCROLL';
		$00B0: Result := 'EM_GETSEL';
		$00B1: Result := 'EM_SETSEL';
		$00B2: Result := 'EM_GETRECT';
		$00B3: Result := 'EM_SETRECT';
		$00B4: Result := 'EM_SETRECTNP';
		$00B5: Result := 'EM_SCROLL';
		$00B6: Result := 'EM_LINESCROLL';
		$00B7: Result := 'EM_SCROLLCARET';
		$00B8: Result := 'EM_GETMODIFY';
		$00B9: Result := 'EM_SETMODIFY';
		$00BA: Result := 'EM_GETLINECOUNT';
		$00BB: Result := 'EM_LINEINDEX';
		$00BC: Result := 'EM_SETHANDLE';
		$00BD: Result := 'EM_GETHANDLE';
		$00BE: Result := 'EM_GETTHUMB';
		$00C1: Result := 'EM_LINELENGTH';
		$00C2: Result := 'EM_REPLACESEL';
		$00C4: Result := 'EM_GETLINE';
		$00C5: Result := 'EM_LIMITTEXT / EM_SETLIMITTEXT';
		$00C6: Result := 'EM_CANUNDO';
		$00C7: Result := 'EM_UNDO';
		$00C8: Result := 'EM_FMTLINES';
		$00C9: Result := 'EM_LINEFROMCHAR';
		$00CB: Result := 'EM_SETTABSTOPS';
		$00CC: Result := 'EM_SETPASSWORDCHAR';
		$00CD: Result := 'EM_EMPTYUNDOBUFFER';
		$00CE: Result := 'EM_GETFIRSTVISIBLELINE';
		$00CF: Result := 'EM_SETREADONLY';
		$00D0: Result := 'EM_SETWORDBREAKPROC';
		$00D1: Result := 'EM_GETWORDBREAKPROC';
		$00D2: Result := 'EM_GETPASSWORDCHAR';
		211: Result := 'EM_SETMARGINS';
		212: Result := 'EM_GETMARGINS';
		213: Result := 'EM_GETLIMITTEXT';
		214: Result := 'EM_POSFROMCHAR';
		215: Result := 'EM_CHARFROMPOS';
		216: Result := 'EM_SETIMESTATUS';
		217: Result := 'EM_GETIMESTATUS';

		224: Result := 'SBM_SETPOS';
		225: Result := 'SBM_GETPOS';
		226: Result := 'SBM_SETRANGE';
		230: Result := 'SBM_SETRANGEREDRAW';
		227: Result := 'SBM_GETRANGE';
		228: Result := 'SBM_ENABLE_ARROWS';
		233: Result := 'SBM_SETSCROLLINFO';
		234: Result := 'SBM_GETSCROLLINFO';

		(WM_USER+0): Result := 'DM_GETDEFID';
		(WM_USER+1): Result := 'DM_SETDEFID';
		(WM_USER+2): Result := 'DM_REPOSITION';
		(WM_USER+100): Result := 'PSM_PAGEINFO';
		(WM_USER+101): Result := 'PSM_SHEETINFO';

		CM_BASE + 0: Result := 'CM_ACTIVATE';
		CM_BASE + 1: Result := 'CM_DEACTIVATE';
		CM_BASE + 2: Result := 'CM_GOTFOCUS';
		CM_BASE + 3: Result := 'CM_LOSTFOCUS';
		CM_BASE + 4: Result := 'CM_CANCELMODE';
		CM_BASE + 5: Result := 'CM_DIALOGKEY';
		CM_BASE + 6: Result := 'CM_DIALOGCHAR';
		CM_BASE + 7: Result := 'CM_FOCUSCHANGED';
		CM_BASE + 8: Result := 'CM_PARENTFONTCHANGED';
		CM_BASE + 9: Result := 'CM_PARENTCOLORCHANGED';
		CM_BASE + 10: Result := 'CM_HITTEST';
		CM_BASE + 11: Result := 'CM_VISIBLECHANGED';
		CM_BASE + 12: Result := 'CM_ENABLEDCHANGED';
		CM_BASE + 13: Result := 'CM_COLORCHANGED';
		CM_BASE + 14: Result := 'CM_FONTCHANGED';
		CM_BASE + 15: Result := 'CM_CURSORCHANGED';
		CM_BASE + 16: Result := 'CM_CTL3DCHANGED';
		CM_BASE + 17: Result := 'CM_PARENTCTL3DCHANGED';
		CM_BASE + 18: Result := 'CM_TEXTCHANGED';
		CM_BASE + 19: Result := 'CM_MOUSEENTER';
		CM_BASE + 20: Result := 'CM_MOUSELEAVE';
		CM_BASE + 21: Result := 'CM_MENUCHANGED';
		CM_BASE + 22: Result := 'CM_APPKEYDOWN';
		CM_BASE + 23: Result := 'CM_APPSYSCOMMAND';
		CM_BASE + 24: Result := 'CM_BUTTONPRESSED';
		CM_BASE + 25: Result := 'CM_SHOWINGCHANGED';
		CM_BASE + 26: Result := 'CM_ENTER';
		CM_BASE + 27: Result := 'CM_EXIT';
		CM_BASE + 28: Result := 'CM_DESIGNHITTEST';
		CM_BASE + 29: Result := 'CM_ICONCHANGED';
		CM_BASE + 30: Result := 'CM_WANTSPECIALKEY';
		CM_BASE + 31: Result := 'CM_INVOKEHELP';
		CM_BASE + 32: Result := 'CM_WINDOWHOOK';
		CM_BASE + 33: Result := 'CM_RELEASE';
		CM_BASE + 34: Result := 'CM_SHOWHINTCHANGED';
		CM_BASE + 35: Result := 'CM_PARENTSHOWHINTCHANGED';
		CM_BASE + 36: Result := 'CM_SYSCOLORCHANGE';
		CM_BASE + 37: Result := 'CM_WININICHANGE';
		CM_BASE + 38: Result := 'CM_FONTCHANGE';
		CM_BASE + 39: Result := 'CM_TIMECHANGE';
		CM_BASE + 40: Result := 'CM_TABSTOPCHANGED';
		CM_BASE + 41: Result := 'CM_UIACTIVATE';
		CM_BASE + 42: Result := 'CM_UIDEACTIVATE';
		CM_BASE + 43: Result := 'CM_DOCWINDOWACTIVATE';
		CM_BASE + 44: Result := 'CM_CONTROLLISTCHANGE';
		CM_BASE + 45: Result := 'CM_GETDATALINK';
		CM_BASE + 46: Result := 'CM_CHILDKEY';
		CM_BASE + 47: Result := 'CM_DRAG';
		CM_BASE + 48: Result := 'CM_HINTSHOW';
		CM_BASE + 49: Result := 'CM_DIALOGHANDLE';
		CM_BASE + 50: Result := 'CM_ISTOOLCONTROL';
		CM_BASE + 51: Result := 'CM_RECREATEWND';
		CM_BASE + 52: Result := 'CM_INVALIDATE';
		CM_BASE + 53: Result := 'CM_SYSFONTCHANGED';
		CM_BASE + 54: Result := 'CM_CONTROLCHANGE';
		CM_BASE + 55: Result := 'CM_CHANGED';
		CM_BASE + 56: Result := 'CM_DOCKCLIENT';
		CM_BASE + 57: Result := 'CM_UNDOCKCLIENT';
		CM_BASE + 58: Result := 'CM_FLOAT';
		CM_BASE + 59: Result := 'CM_BORDERCHANGED';
		CM_BASE + 60: Result := 'CM_BIDIMODECHANGED';
		CM_BASE + 61: Result := 'CM_PARENTBIDIMODECHANGED';
		CM_BASE + 62: Result := 'CM_ALLCHILDRENFLIPPED';
		CM_BASE + 63: Result := 'CM_ACTIONUPDATE';
		CM_BASE + 64: Result := 'CM_ACTIONEXECUTE';
		CM_BASE + 65: Result := 'CM_HINTSHOWPAUSE';
		CM_BASE + 66: Result := 'CM_DOCKNOTIFICATION';
		CM_BASE + 67: Result := 'CM_MOUSEWHEEL';

		CN_BASE + WM_CHARTOITEM: Result := 'CN_CHARTOITEM';
		CN_BASE + WM_COMMAND: Result := 'CN_COMMAND';
		CN_BASE + WM_COMPAREITEM: Result := 'CN_COMPAREITEM';
		CN_BASE + WM_CTLCOLORBTN: Result := 'CN_CTLCOLORBTN';
		CN_BASE + WM_CTLCOLORDLG: Result := 'CN_CTLCOLORDLG';
		CN_BASE + WM_CTLCOLOREDIT: Result := 'CN_CTLCOLOREDIT';
		CN_BASE + WM_CTLCOLORLISTBOX: Result := 'CN_CTLCOLORLISTBOX';
		CN_BASE + WM_CTLCOLORMSGBOX: Result := 'CN_CTLCOLORMSGBOX';
		CN_BASE + WM_CTLCOLORSCROLLBAR: Result := 'CN_CTLCOLORSCROLLBAR';
		CN_BASE + WM_CTLCOLORSTATIC: Result := 'CN_CTLCOLORSTATIC';
		CN_BASE + WM_DELETEITEM: Result := 'CN_DELETEITEM';
		CN_BASE + WM_DRAWITEM: Result := 'CN_DRAWITEM';
		CN_BASE + WM_HSCROLL: Result := 'CN_HSCROLL';
		CN_BASE + WM_MEASUREITEM: Result := 'CN_MEASUREITEM';
		CN_BASE + WM_PARENTNOTIFY: Result := 'CN_PARENTNOTIFY';
		CN_BASE + WM_VKEYTOITEM: Result := 'CN_VKEYTOITEM';
		CN_BASE + WM_VSCROLL: Result := 'CN_VSCROLL';
		CN_BASE + WM_KEYDOWN: Result := 'CN_KEYDOWN';
		CN_BASE + WM_KEYUP: Result := 'CN_KEYUP';
		CN_BASE + WM_CHAR: Result := 'CN_CHAR';
		CN_BASE + WM_SYSKEYDOWN: Result := 'CN_SYSKEYDOWN';
		CN_BASE + WM_SYSCHAR: Result := 'CN_SYSCHAR';
		CN_BASE + WM_NOTIFY: Result := 'CN_NOTIFY';
	else
		Result := 'Unknown';
	end;
end;

procedure TMessageInterceptor.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (AComponent = FControl) and (Operation = opRemove) then
    Control := nil;
end;

procedure TMessageInterceptor.SetActive(const Value: boolean);
begin
	if csLoading in ComponentState then
  begin
		FActive := Value;
		Exit;
	end;
	if FActive <> Value then
  begin
		if not Assigned(FControl) then
    begin
			ShowMessage('Control is not assigned');
			Exit;
		end;
		if FActive then
      FControl.WindowProc := FDefaultWindowProc;
		FActive := Value;
		if FActive then
      AssignWindowProc;
	end;
end;

procedure TMessageInterceptor.SetControl(const Value: TControl);
begin
	if FControl <> Value then
  begin
		if not (csLoading in ComponentState) then
    begin
			if Assigned(FControl) and FActive then
        FControl.WindowProc := FDefaultWindowProc;
		end;
		FControl := Value;
    if Assigned(FControl) then
    begin
      FControl.FreeNotification(Self);
      if Active then
        AssignWindowProc;
    end;
	end;
end;

procedure TMessageInterceptor._WindowProc(var Msg: TMessage);
var
	bContinue: boolean;
begin
	bContinue := True;
	if Assigned(FBeforeMessage) then
		FBeforeMessage(Self, Msg, bContinue);
	if bContinue then
	begin
		FDefaultWindowProc(Msg);
		if Assigned(FAfterMessage) then
			FAfterMessage(Self, Msg);
	end;
end;

end.
