<?php
/*************************************************************************************
   Copyright notice
   
   (c) 2002-2003 Oliver Georgi (oliver@phpwcms.de) // All rights reserved.
 
   This script is part of PHPWCMS. The PHPWCMS web content management system is
   free software; you can redistribute it and/or modify it under the terms of
   the GNU General Public License as published by the Free Software Foundation;
   either version 2 of the License, or (at your option) any later version.
  
   The GNU General Public License can be found at http://www.gnu.org/copyleft/gpl.html
   A copy is found in the textfile GPL.txt and important notices to the license 
   from the author is found in LICENSE.txt distributed with these scripts.
  
   This script is distributed in the hope that it will be useful, but WITHOUT ANY 
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
   This copyright notice MUST APPEAR in all copies of the script!
*************************************************************************************/

/***********************************************************************************************************
| imagick_converting (c) 2003 by Oliver Georgi, http://www.webverbund.de
|-----------------------------------------------------------------------------------------------------------
| STRING	$img		= name of the image which should be converted
| STRING	$ext		= extension of target image ("tif", "png", "jpg", "gif" ...)
| STRING	$ttxt		= text added to the name of image ("_thumb" = name_thumb.ext)
|						  use generic_string($length, $i=0) to add a random string like 'a0Bc9Kuc3...'
|						  like: "_".generic_string(10, 0)
| STRING	$imgdir		= where is the image located which should be converted
|						  (if in use finish the end by '/' like '/root/my_image_dir/')
| STRING	$tdir		= where should the converted image be saved
|						  (if in use finish the end by '/' like '/root/my_thumbs_dir/')
| INT		$mw			= max width of converted image (only if larger original width)
| INT		$mh			= max height of converted image (only if larger original height)
| INT		$qual		= image compression quality default = 0 means ImageMagick default
| INT/STR	$prfl		= 0 = all profile informations will be deleted
|						  or ImageMagick Profile attribut string like -profile...
| INT		$dens		= density, default 72, if set to 0 then as is		
| STRING	$addcmd		= additinal ImageMagick Convert command
************************************************************************************************************/

if(!IMAGICK_ON) include PHPWCMS_ROOT."/include/inc_lib/ext/hft_image/hft_image.php";

function imagick_converting ($img,$ext="jpg",$ttxt="",$imgdir="",$tdir="",$mw=150,$mh=150,$qual=75,$prfl=0,$dens=72,$addcmd="") {

	$imagick["error"]			= 0;
	$imagick["max_width"]		= intval($mw); if(!$imagick["max_width"]) $imagick["max_width"] = "";
	$imagick["max_height"]		= intval($mh); if(!$imagick["max_height"]) $imagick["max_height"] = "";
	$imagick["image_name"]		= $img;
	$imagick["target_ext"]		= (IMAGICK_ON) ? $ext : "jpg"; //If GD is on -> only use jpg as result
	$imagick["thumb_text"]		= $ttxt;
	$imagick["image_dir"]		= $imgdir;
	$imagick["thumb_dir"]		= $tdir;
	
	$imagick["image_thumb_name"] = cut_ext($imagick["image_name"]).$imagick["thumb_text"];
	$imagick["image_thumb_name"] = str_replace(" ", "_", $imagick["image_thumb_name"]);
	$imagick["image_thumb_name"] .= ($imagick["target_ext"]) ? (".".$imagick["target_ext"]) : (".".which_ext($imagick["image_name"]));
	
	if(IMAGICK_ON) {
		// If ImageMagick should be used
		$imagick["command"]  = IMAGICK_PATH."convert ";
		switch($imagick["target_ext"]) {
			case "jpg":  $imagick["command"] .= "-colorspace RGB -type TrueColor -sharpen 1x10 "; break; // -flatten -sharpen 3x50
			case "jpeg": $imagick["command"] .= "-colorspace RGB -type TrueColor -sharpen 1x10 "; break; // -flatten -sharpen 3x50
			case "gif":  $imagick["command"] .= "-colors 128 "; break;
			case "png":  $imagick["command"] .= "-colors 128 "; break;
		}
		if($imagick["max_width"] || $imagick["max_height"]) $imagick["command"] .= "-geometry \"".$imagick["max_width"]."x".$imagick["max_height"].">\" ";
		if($qual) $imagick["command"] .= "-quality ".$qual." ";
		if($dens) $imagick["command"] .= "-density ".$dens."x".$dens." ";
		if($addcmd) $imagick["command"] .= $addcmd." ";
		$imagick["command"] .= "\"".$imagick["image_dir"].$imagick["image_name"]."[0]\" ";
		if(!$prfl) { $imagick["command"] .= "+profile \"*\" "; } else { $imagick["command"] .= $prfl." "; }
		$imagick["command"] .= "\"".$imagick["thumb_dir"].$imagick["image_thumb_name"]."\" ";
		
		@exec($imagick["command"], $imagick_return);
		if ($imagick_return) $imagick["error"] = $imagick_return;
		
	} else {
		// use GD function
		$check_image = getimagesize($imagick["image_dir"].$imagick["image_name"]);
		if($check_image) {
		
			if($imagick["max_width"] && $check_image[0] < $imagick["max_width"]) $imagick["max_width"] = $check_image[0];
			if($imagick["max_height"] && $check_image[1] < $imagick["max_height"]) $imagick["max_height"] = $check_image[1];		
			if(!$imagick["max_width"]) $imagick["max_width"] = "*";
			if(!$imagick["max_height"]) $imagick["max_height"] = "*";
		
			$image = new hft_image($imagick["image_dir"].$imagick["image_name"]); //given original image
			$image->set_parameters($qual, GD2_ON);
			$image->resize($imagick["max_width"], $imagick["max_height"], '-');
			$image->output_resized($imagick["thumb_dir"].$imagick["image_thumb_name"], "JPEG");
			if(!file_exists($imagick["thumb_dir"].$imagick["image_thumb_name"])) $imagick["error"] = "GD image creation failed.";
		} else {
			 $imagick["error"] = "GD source image error. Maybe not exists.";
		}
	}

	return $imagick;
}

?>