{*******************************************************}
{                                                       }
{         TreeView enhancement with wallpaper,          }
{         CheckBoxes and                                }
{         string data assigned to each node             }
{                                                       }
{         Copyright (c) FINANCES Ltd                    }
{         2003   ver. 1                                 }
{                                                       }
{         www.finances-ltd.narod.ru                     }
{         finances-ltd@narod.ru                         }
{                                                       }
{*******************************************************}

unit zTreeView;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, ComCtrls, CommCtrl;

const
  {ITEM_ENABLED        = integer(true);
  ITEM_DISABLED       = integer(false);
  ncsUndefined        = -1;
  ncsUnChecked        = 1;
  ncsChecked          = 2;
  ncsPartChecked      = 3;}

  TVS_NOTOOLTIPS        = $0080;
  TVS_CHECKBOXES        = $0100;
  TVS_TRACKSELECT       = $0200;

type
  FOnTNDrawEvent       =  procedure (Sender: TObject; Message: TWMPaint ) of object;

  TzzCustomTreeView = class(TCustomTreeView)
  private
    FCanvas		: TCanvas;
    FWallpaper		: TBitmap;
    FBoldSelection      : boolean;
    FHotTrack           : Boolean;
    FCheckboxes         : Boolean;
    FToolTips           : Boolean;
    FMask               : TBitmap;
    FOnDraw             : FOnTNDrawEvent;
    FOnEndEdit          : TNotifyEvent;
    fPaintingNow	: boolean;
    isEditing_          : boolean;
    FCutNode: TTreeNode;
    //FOnCollapsed: TTVExpandedEvent;
    //FOnCollapsing: TTVCollapsingEvent;
    //FOnExpanded: TTVExpandedEvent;
    //FOnExpanding: TTVExpandingEvent;
    procedure Paint(var Message: TWMPaint); message WM_PAINT;
    procedure SetWallpaper( Value: TBitmap );
    function GetWallpaper: TBitmap;
    procedure SetBoldSelection( Value: boolean );
    procedure SetHotTrack(Value: Boolean);
    procedure SetCheckboxes(Value: Boolean);
    procedure SetToolTips(Value: Boolean);
//    procedure DoDragOver(Source: TDragObject; X, Y: Integer; CanDrop: Boolean); override;
    //procedure SetExpanded(Node: TTreeNode);
    procedure CMDesignHitTest(var Message: TCMDesignHitTest); message CM_DESIGNHITTEST;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
//    procedure CNHScroll(var Message: TWMHScroll); message CN_HSCROLL;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    {property OnCollapsed: TTVExpandedEvent read FOnCollapsed write FOnCollapsed;
    property OnCollapsing: TTVCollapsingEvent read FOnCollapsing write FOnCollapsing;
    property OnExpanding: TTVExpandingEvent read FOnExpanding write FOnExpanding;
    property OnExpanded: TTVExpandedEvent read FOnExpanded write SetExpanded;}
//    procedure UpdateFlatScrollBar; virtual;
  public
    fDefaultPainting: boolean;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property Canvas: TCanvas read FCanvas;
    procedure SetNodeBoldState( Node: TTreeNode; fBold: boolean );
    procedure SetNodeState_( ItemID: HTreeItem; stateMask, State: UINT );

    function CanPaste(Node: TTreeNode = nil): Boolean;
    function Cut(Node: TTreeNode = nil): Boolean;
    function Paste(Node: TTreeNode = nil): Boolean;

    function AddObject(Sibling: TTreeNode; const S: string; Ptr: Pointer): TTreeNode; overload;
    function AddObject(Node: TTreeNode; const S: string; DataString: string): TTreeNode; overload;
    function AddChildObject(Parent: TTreeNode; const S: string; Ptr: Pointer): TTreeNode; overload;
    function AddChildObject(Node: TTreeNode; const S: string; DataString: string): TTreeNode; overload;
    function AddChildObjectFirst(Parent: TTreeNode; const S: string; Ptr: Pointer): TTreeNode; overload;
    function AddChildObjectFirst(Parent: TTreeNode; const S: string; DataString: string): TTreeNode; overload;

    procedure SetNodeData(Data: String; Node: TtreeNode = nil);
    function GetNodeData(Node: TtreeNode = nil):string;
    function IndexOfData(Data: string): Integer;

    function MoveUp(Node: TTreeNode = nil):boolean;
    function MoveDown(Node: TTreeNode = nil):boolean;
    function CanMoveDown(Node: TTreeNode = nil): Integer;

    procedure SaveToStreamEx(Stream:Tstream);
    procedure LoadFromStreamEx(Stream:Tstream);

    procedure SaveToFileEx(const FileName: string);
    procedure LoadFromFileEx(const FileName: string);

    procedure SaveToStrings(List: TStrings; WithData: Boolean = True);
    procedure LoadFromStrings(List: TStrings);

    function IsChecked(Node: TTreeNode): Boolean;
    procedure SetChecked(Node: TTreeNode; Checked: Boolean);
    function CheckedCount: Integer;

  protected
    property Wallpaper: TBitmap read GetWallpaper write SetWallpaper;
    property BoldSelection: boolean read FBoldSelection write SetBoldSelection;
    property HotTrack: Boolean read FHotTrack write SetHotTrack default False;
    property CheckBoxes: Boolean read FCheckboxes write SetCheckboxes default False;
    property ToolTips: Boolean read FToolTips write SetToolTips default True;
    property OnDraw: FOnTNDrawEvent read FOnDraw write FOnDraw;
    property OnEndEdit: TNotifyEvent read FOnEndEdit write FOnEndEdit;
  end;

  TzTreeView = class(TzzCustomTreeView)
  published
    property Anchors;
    property ShowButtons;
    property BorderStyle;
    property ChangeDelay;
    property DragCursor;
    property ShowLines;
    property ShowRoot;
    property ReadOnly;
    property OnCustomDrawItem;
    property RightClickSelect;
    property DragMode;
    property HideSelection;
    property Indent;
    property Items;
    property MultiSelect;
    property MultiSelectStyle;
    property OnEditing;
    property OnEdited;
    property OnExpanding;
    property OnExpanded;
    property OnCollapsing;
    property OnCompare;
    property OnCollapsed;
    property OnChanging;
    property OnChange;
    property OnDeletion;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property Align;
    property Enabled;
    property Font;
    property Color;
    property ParentColor default False;
    property ParentCtl3D;
    property Ctl3D;
    property SortType;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnClick;
    property OnEnter;
    property OnExit;
    property OnDragDrop;
    property OnDragOver;
    property OnStartDrag;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnDblClick;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property PopupMenu;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property Images;
    property StateImages;
    property Wallpaper;
    property BoldSelection;
    property HotTrack;
    property CheckBoxes;
    property ToolTips;
    property OnDraw;
    property OnEndEdit;
  end;

type Pdatatypestring=^string;
const  sInvalidLevelEx = 'Invalid level (%d) for item "%s"';
       TabChar = #9;
       EndOfLine = #13#10;
       DividerChar = '^';
       DividerChar2 = '~';

       TVIS_CHECKED  =  $2000;

procedure Register;

implementation
//{$R glTView.res}
//uses glUtils;

procedure Register;
begin
  RegisterComponents('zPage', [TzTreeView]);
end;

function GetBufStart(Buffer: PChar; var Level: Integer): PChar;
begin
  Level := 0;
  while Buffer^ in [' ', #9] do
  begin
    Inc(Buffer);
    Inc(Level);
  end;
  Result := Buffer;
end;

procedure TreeViewErrorFmt(const Msg: string; Format: array of const);
begin
  raise ETreeViewError.CreateFmt(Msg, Format);
end;

constructor TzzCustomTreeView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCanvas := TControlCanvas.Create;
  TControlCanvas(FCanvas).Control := Self;//...i can draw now! :)
  FHotTrack := False;
  FCheckboxes := False;
  FToolTips := True;
  FMask := TBitmap.Create;
  if csDesigning in ComponentState then FWallpaper := TBitmap.Create;
end;

destructor TzzCustomTreeView.Destroy;
begin
  FCanvas.Free;
  if Assigned(FWallpaper) then FWallpaper.Free;
  if Assigned(FMask) then FMask.Free;
  inherited Destroy;
end;

procedure TzzCustomTreeView.Assign(Source: TPersistent);
begin
  if Source is TzzCustomTreeView then
  begin
    Width:=TzzCustomTreeView(Source).Width;
    Height:=TzzCustomTreeView(Source).Height;
    Align:=TzzCustomTreeView(Source).Align;
    Parent:=TzzCustomTreeView(Source).Parent;
    Items.Assign(TzzCustomTreeView(Source).Items);
  end;
end;

procedure TzzCustomTreeView.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    if FHotTrack then Style:= Style or TVS_TRACKSELECT;
    if FCheckBoxes then Style:= Style or TVS_CHECKBOXES;
    if not FToolTips then Style:= Style or TVS_NOTOOLTIPS;
  end;
end;

procedure TzzCustomTreeView.CMDesignHitTest(var Message: TCMDesignHitTest);
begin
  if htOnButton in GetHitTestInfoAt(Message.XPos, Message.YPos) then
    Message.Result := 1 //...expanding in Design Time
  else
    Message.Result := 0;
end;

(*procedure TzzCustomTreeView.UpdateFlatScrollBar;
var
  MinPos, MaxPos: integer;
begin
  {$IFDEF GLVER_D5}
  try
    InitializeFlatSB(Handle);
    FlatSB_SetScrollProp(handle,WSB_PROP_VSTYLE, FSB_ENCARTA_MODE, false);
    FlatSB_SetScrollProp(handle,WSB_PROP_HSTYLE, FSB_ENCARTA_MODE, false);

    GetScrollRange(Handle, SB_VERT, MinPos, MaxPos);

    FlatSB_SetScrollRange(handle, 1, MinPos, MaxPos, false);
  except end;                                                                  
  {$ENDIF}
end;*)

procedure TzzCustomTreeView.Paint(var Message: TWMPaint);
var
  x,y,IWidth,IHeight: integer;
  r: TRect;
  Bmp, Bmp2: TBitmap;
  mask, OldMask: HBITMAP;
  MaskDC: HDC;
  OldBkColor: COLORREF;
begin
  if csDestroying in ComponentState then Exit;
  try
    if not Assigned(FWallpaper)or(FWallpaper.handle=0) then
    begin
      inherited;
      if Assigned(FOnDraw) then FOnDraw( self, Message );
      exit;
    end;

    fPaintingNow := true;
    r:=GetClientRect;
    IWidth := r.right-r.left; IHeight := r.bottom-r.top;
    Bmp := TBitmap.Create;
    Bmp2 := TBitmap.Create;
    mask := CreateBitmap( IWidth, IHeight, 1, 1, nil );
    MaskDC := CreateCompatibleDC( Canvas.handle );
    OldMask := SelectObject( MaskDC, Mask );

    bmp.Width := IWidth; bmp.Height := IHeight;
    bmp2.Width := IWidth; bmp2.Height := IHeight;
    x := 0; y := 0;
    while x < r.Right-r.Left do begin
      while y < IHeight do begin
        BitBlt( bmp.Canvas.Handle, x, y, FWallpaper.Width, FWallpaper.Height, FWallpaper.Canvas.Handle, 0,0, SRCCOPY );
        Inc(y, FWallpaper.Height);
      end;
      Inc(x, FWallpaper.Width);
      y:=0;
    end;

    inherited;
  //  BitBlt( bmp.Canvas.Handle, 0, 0, IWidth, IHeight, FWallpaper.Canvas.Handle, 0,0, SRCCOPY );
    //SetBkMode(Canvas.handle,TRANSPARENT);
    //InvalidateRect(handle,@r,false);
  //  inherited;
  //  bmp2.Width := Width; bmp2.Height := Height;
  //  self.PaintWindow(bmp2.Canvas.Handle);
  //exit;
    BitBlt( bmp2.Canvas.Handle, 0, 0, IWidth, IHeight, Canvas.handle, 0,0, SRCCOPY );

    OldBkColor := SetBkColor( bmp2.Canvas.handle, ColorToRGB(Color) );
    BitBlt( MaskDC, 0, 0, IWidth, IHeight, bmp2.Canvas.handle, 0,0, SRCCOPY );
    SetBkColor( bmp2.Canvas.handle, OldBkColor );

    BitBlt( bmp.Canvas.Handle, 0, 0, IWidth, IHeight, MaskDC, 0,0, SRCAND );

    {put mask on canvas to change background color to white}
    if Color<>clWhite then BitBlt( bmp2.Canvas.Handle, 0, 0, IWidth, IHeight, MaskDC, 0,0, SRCPAINT );

    BitBlt( MaskDC, 0, 0, IWidth, IHeight, MaskDC, 0, 0, NOTSRCCOPY);
    BitBlt( bmp.Canvas.Handle, 0, 0, IWidth, IHeight, MaskDC, 0,0, SRCPAINT );

    BitBlt( bmp2.Canvas.handle, 0, 0, IWidth, IHeight, bmp.Canvas.Handle, 0, 0, SRCAND );
    BitBlt( Canvas.handle, 0, 0, IWidth, IHeight, bmp2.Canvas.Handle, 0, 0, SRCCOPY );

    DeleteObject( SelectObject(MaskDC, OldMask) );
    DeleteDC( MaskDC );
    Bmp.free;
    Bmp2.free;
    fPaintingNow := false;
    if Assigned(FOnDraw) then FOnDraw( self, Message );

  finally
  end;
end;

procedure TzzCustomTreeView.CNNotify(var Message: TWMNotify);
begin
  with Message.NMHdr^ do
    case code of
      TVN_ENDLABELEDIT:
        begin
          if Assigned( OnEndEdit ) then  OnEndEdit(self);
          isEditing_ := false;
        end;
      TVN_SELCHANGING:
        with PNMTreeView(Pointer(Message.NMHdr))^ do
        begin
          if FBoldSelection then
          begin
            SetNodeState_(itemOld.hItem, TVIS_BOLD, 0);
            SetNodeState_(itemNew.hItem, TVIS_BOLD, TVIS_BOLD);
          end;
          inherited;
          exit;
        end;
        TVN_BEGINLABELEDIT: isEditing_ := true;
//      TVN_SELCHANGED:
//        with PNMTreeView(Pointer(Message.NMHdr))^ do
//          inherited;
          //Change(GetNodeFromItem(itemNew));
    end;
  inherited;
end;

procedure TzzCustomTreeView.SetNodeBoldState( Node: TTreeNode; fBold: boolean );
begin
  if fBold then SetNodeState_( Node.ItemID, TVIS_BOLD, TVIS_BOLD )
           else SetNodeState_( Node.ItemID, TVIS_BOLD, 0 )
end;

procedure TzzCustomTreeView.SetNodeState_( ItemID: HTreeItem; stateMask, State: UINT );
var
  tvi: TTVItem;
begin
  FillChar( tvi, Sizeof(tvi), 0 );
  tvi.hItem := ItemID;
  tvi.mask := TVIF_STATE;
  tvi.stateMask := stateMask;//TVIS_DROPHILITED, TVIS_BOLD;
  tvi.state := State;
  TreeView_SetItem( Handle, tvi );
end;

procedure TzzCustomTreeView.SetCheckboxes(Value: Boolean);
begin
  if FCheckboxes = Value then exit;
  FCheckboxes := Value; RecreateWnd;
end;

procedure TzzCustomTreeView.SetHotTrack(Value: Boolean);
begin
  if FHotTrack = Value then exit;
  FHotTrack := Value; RecreateWnd;
end;

procedure TzzCustomTreeView.SetToolTips(Value: Boolean);
begin
  if FToolTips = Value then exit;
  FToolTips := Value;
  RecreateWnd;
end;

//---------------------------------------------propeties
procedure TzzCustomTreeView.SetWallpaper( Value: TBitmap );
begin
  if Assigned(FWallpaper) then FWallpaper.Free;
  FWallpaper := TBitmap.Create;
  FWallpaper.Assign(Value); Invalidate;
end;

function TzzCustomTreeView.GetWallpaper: TBitmap;
begin
  if not Assigned(FWallpaper) then FWallpaper := TBitmap.Create;
  Result := FWallpaper;
end;

procedure TzzCustomTreeView.SetBoldSelection( Value: boolean );
begin
  FBoldSelection := Value;
  if Assigned(Selected) then SetNodeBoldState( Selected, FBoldSelection );
end;

function TzzCustomTreeView.CanPaste(Node: TTreeNode = nil): Boolean;
  function _CanPaste(Node: TTreeNode): Boolean;
  var ParentItem: TTreeNode;
  begin
    Result := False;
    if Assigned(FCutNode) and Assigned(Node) then begin
      if (FCutNode = Node) or (FCutNode = Node.Parent) or
            (FCutNode.Parent = Node) then Exit;

      if FCutNode.Level < Node.Level then begin
        ParentItem := Node.Parent;
        while ParentItem.Level > FCutNode.Level do begin
          ParentItem := ParentItem.Parent;
          if ParentItem = FCutNode then Exit;
        end;
      end;
      Result := True;
    end;
  end;
begin
  if not Assigned(Node) then
    Result := _CanPaste(Selected)
  else Result := _CanPaste(Node)
end;

function TzzCustomTreeView.Cut(Node: TTreeNode = nil): Boolean;
begin
  if Assigned(Node) then begin
    FCutNode := Node;
    Result := FCutNode.Cut;
  end
  else if Assigned(Selected) then begin
    FCutNode := Selected;
    Result := FCutNode.Cut;
  end
  else Result := False;
end;

function TzzCustomTreeView.Paste(Node: TTreeNode = nil): Boolean;
  function _Paste: Boolean;
  begin
    if CanPaste(Node) then begin
      FCutNode.MoveTo(Node, naAddChild);
      Selected := FCutNode;
      Selected.MakeVisible;
      Result := True;
    end
    else Result := False;
  end;
begin
  Result := False;
  if not Assigned(Node) then
    Node := Selected;
  if not Assigned(Node) then Exit;
  Result := _Paste;
end;

function TzzCustomTreeView.AddObject(Sibling: TTreeNode; const S: string;
  Ptr: Pointer): TTreeNode;
begin
  Result := Items.AddNode(nil, Sibling, S, Ptr, naAdd);
end;

function TzzCustomTreeView.AddObject(Node: TTreeNode; const S: string; DataString:string): TTreeNode;
var ptr:Pdatatypestring;
begin
  if Length(s) = 0 then
    Result := Items.AddObject(Node,s,nil)
  else begin
    new(ptr);
    ptr^ := Datastring;
    Result := Items.AddObject(Node,s,ptr);
  end;
end;

function TzzCustomTreeView.AddChildObject(Parent: TTreeNode; const S: string;
  Ptr: Pointer): TTreeNode;
begin
  Result := Items.AddNode(nil, Parent, S, Ptr, naAddChild);
end;

function TzzCustomTreeView.AddChildObject(Node: TTreeNode; const S: string; DataString:string): TTreeNode;
var ptr:Pdatatypestring;
begin
  if Length(s) = 0 then
    Result := Items.AddChildObject(Node,s,nil)
  else begin
    new(ptr);
    ptr^ := Datastring;
    Result := Items.AddChildObject(Node, s, ptr);
  end;
end;

function TzzCustomTreeView.AddChildObjectFirst(Parent: TTreeNode; const S: string; Ptr: Pointer): TTreeNode;
begin
  Result := Items.AddNode(nil, Parent, S, Ptr, naAddChildFirst);
end;

function TzzCustomTreeView.AddChildObjectFirst(Parent: TTreeNode; const S: string; DataString: string): TTreeNode;
var ptr:Pdatatypestring;
begin
  if Length(s) = 0 then
    Result := Items.AddNode(nil, Parent, S, nil, naAddChildFirst)
  else begin
    new(ptr);
    ptr^ := Datastring;
    Result := Items.AddNode(nil, Parent, S, Ptr, naAddChildFirst);
  end;
end;

procedure TzzCustomTreeView.SetNodeData(Data: String; Node: TtreeNode = nil);
var ptr:Pdatatypestring;
begin
  if not Assigned(Node) then begin
    if Assigned(Selected) then
      Node := Selected
    else Exit;
  end;

  if length(data)= 0 then
    Node.Data := nil
  else begin
    new(ptr);
    ptr^ := Data;
    Node.Data := ptr;
  end;
end;

Function TzzCustomTreeView.GetNodeData(Node: TtreeNode = nil):string;
begin
  Result := '';
  if not Assigned(Node) then begin
    if Assigned(Selected) then
      Node := Selected
    else Exit;
  end;
  if (Node.Data <> nil) then
    Result := string(Node.Data^);
end;

function TzzCustomTreeView.IndexOfData(Data: string): Integer;
var i: Integer;
begin
  Result := -1;
  for i := 0 to Pred(Items.Count) do
    if GetNodeData(Items[i]) = Data then begin
      Result := i;
      Break;
    end;
end;

function TzzCustomTreeView.MoveUp(Node: TTreeNode = nil): Boolean;
var AnItem: TTreeNode;
begin
  Result := False;
  if not Assigned(Node) then begin
    if Assigned(Selected) then
      Node := Selected
    else Exit;
  end;

  AnItem := Items[Node.AbsoluteIndex - 1];
  while not AnItem.IsVisible do
    AnItem := AnItem.Parent;
  if (AnItem.Level = Node .Level) or (AnItem = Node.Parent) then
    Node.MoveTo(AnItem, naInsert)
  else Node.MoveTo(AnItem.Parent, naAddChild);
  Result := True;
end;

function TzzCustomTreeView.MoveDown(Node: TTreeNode = nil): Boolean;
var AnItem: TTreeNode;
    i: Integer;
begin
  Result := False;
  if not Assigned(Node) then begin
    if Assigned(Selected) then
      Node := Selected
    else Exit;
  end;

  i := CanMoveDown(Node);
  if i = -1 then Exit;
  AnItem := Items[i];
  if AnItem.HasChildren and AnItem.Expanded then
    Node.MoveTo(AnItem, naAddChildFirst)
  else if (AnItem.Level = Node .Level) then
    AnItem.MoveTo(Node, naInsert)
  else Node.MoveTo(AnItem, naInsert);
  Result := True;
end;

function TzzCustomTreeView.CanMoveDown(Node: TTreeNode = nil): Integer;
var i: Integer;
begin
  Result := -1;
  if not Assigned(Node) then begin
    if Assigned(Selected) then
      Node := Selected
    else Exit;
  end;

  i := Node.AbsoluteIndex + 1;
  while (i < Items.Count) and (not Items[i].IsVisible or (Node.Level < Items[i].Level)) do
    Inc(i);
  if i = Items.Count then Exit;
  Result := i;
end;

procedure TzzCustomTreeView.SaveToStrings(List: TStrings; WithData: Boolean = True);
const
  TabChar = #9;
  EndOfLine = #13#10;
var
  i: Integer;
  ANode: TTreeNode;
  NodeStr: string;
begin
  List.Clear;
  if Items.Count > 0 then begin
    ANode := Items[0];
    while ANode <> nil do
    begin
      NodeStr := '';
      for i := 0 to ANode.Level - 1 do NodeStr := NodeStr + TabChar;
      if WithData then
        List.Add(NodeStr + ANode.Text + '=' + GetNodeData(ANode))
      else List.Add(NodeStr + ANode.Text + EndOfLine);
      ANode := ANode.GetNext;
    end;
    List.Text := Trim(List.Text);
  end;
end;

procedure TzzCustomTreeView.LoadFromStrings(List: TStrings);

  function GetBufStart(Buffer: PChar; var Level: Integer): PChar;
  begin
    Level := 0;
    while Buffer^ in [' ', #9] do
    begin
      Inc(Buffer);
      Inc(Level);
    end;
    Result := Buffer;
  end;

var
  ANode, NextNode: TTreeNode;
  ALevel, i: Integer;
  aText, aData: string;
begin
  Items.BeginUpdate;
  try
    try
      Items.Clear;
      ANode := nil;
      for i := 0 to List.Count - 1 do
      begin
        aText := TrimLeft(GetBufStart(PChar(List[i]), ALevel));
        if List.Names[i] <> '' then begin
          aData := List.ValueFromIndex[i];
          aText := TrimLeft(List.Names[i]);
        end
        else aData := '';

        if ANode = nil then
          ANode := Items.AddChild(nil, aText)
        else if ANode.Level = ALevel then
          ANode := Items.AddChild(ANode.Parent, aText)
        else if ANode.Level = (ALevel - 1) then
          ANode := Items.AddChild(ANode, aText)
        else if ANode.Level > ALevel then
        begin
          NextNode := ANode.Parent;
          while NextNode.Level > ALevel do
            NextNode := NextNode.Parent;
          ANode := Items.AddChild(NextNode.Parent, aText);
        end
        else raise ERangeError.CreateFmt(sInvalidLevelEx, [ALevel, aText]);

        if aData <> '' then
          SetNodeData(aData, ANode);
      end;
    finally
      Items.EndUpdate;
    end;
  except
    Invalidate;
    raise;
  end;
end;

procedure TzzCustomTreeView.SaveToStreamEx(Stream:Tstream);
var
  i: Integer;
  ANode: TTreeNode;
  NodeStr: string;
begin

if Items.Count > 0 then
  begin
    ANode := topitem;
    while ANode <> nil do
    begin
      NodeStr := '';
      for i := 0 to ANode.Level - 1 do NodeStr := NodeStr + TabChar;
      if anode.data <> nil then
         NodeStr := NodeStr + ANode.Text+
         Dividerchar2+inttostr(Anode.ImageIndex)+ {writing imageindex}
         dividerchar+string(anode.data^)+EndOfLine{writing StringData}

         else NodeStr := NodeStr + ANode.Text+Dividerchar2+
         inttostr(Anode.ImageIndex)+EndOfLine;

      Stream.Write(Pointer(NodeStr)^, Length(NodeStr));
      ANode := ANode.GetNext;
    end; {while}
  end; {if}
end;

procedure TzzCustomTreeView.LoadFromStreamEx(Stream:Tstream);
var
  List : TStringList;
  ANode, NextNode : TTreeNode;
  ALevel, i : Integer;
  CurrStr : string;
  AImageIndex:Integer;
  DataString : String;
  DataPtr : Pdatatypestring;
begin
  DataPtr := Nil;
  List := TStringList.Create;
  Items.beginupdate;
 try
    try
      Items.Clear;
      List.LoadFromStream(Stream);
      ANode := nil;
      for i := 0 to List.Count - 1 do
       begin
         CurrStr := GetBufStart(PChar(List[i]), ALevel);
          if pos(DividerChar,Currstr)<> 0 then
            begin
            Datastring := copy(CurrStr,pos(DividerChar, CurrStr)+1,
                               length(CurrStr)-pos(DividerChar, CurrStr));
                                setlength(CurrStr,pos(DividerChar, CurrStr)-1);
            new(dataptr);
            dataptr^ := Datastring;        {getting datastring}
            end
              else datastring := '';


           if pos(DividerChar2,Currstr)<> 0 then
            begin
             AImageIndex := strtoint(copy(CurrStr,pos(DividerChar2, CurrStr)+1,
                                    length(CurrStr)-pos(DividerChar2, CurrStr) ) );
             setlength(CurrStr,pos(DividerChar2, CurrStr)-1); {geting ImageIndex}
            end
              else AImageIndex := 0;


         if ANode = nil then
                         begin
                          ANode := Items.AddChild(nil, CurrStr);
                          if Datastring <> '' then Anode.data:=DataPtr;
                         end
            else

              if ANode.Level = ALevel then
                 begin
                  ANode := Items.AddChild(ANode.Parent, CurrStr);
                  if Datastring <> '' then Anode.data := DataPtr;
                 end
               else
                   if ANode.Level = (ALevel - 1) then
                       begin
                        ANode := Items.AddChild(ANode, CurrStr);
                        if Datastring <> '' then Anode.data:=DataPtr;
                       end
                   else if ANode.Level > ALevel then
                            begin
                              NextNode := ANode.Parent;
                              while NextNode.Level > ALevel do NextNode := NextNode.Parent;
                              ANode := Items.AddChild(NextNode.Parent, CurrStr);
                              if Datastring <> '' then Anode.data:=DataPtr;
                            end
                        else TreeViewErrorFmt(sInvalidLevelEx, [ALevel, CurrStr]);

      Anode.ImageIndex := AimageIndex;
      Anode.SelectedIndex := AimageIndex;
      end; {for}
    finally
    Items.EndUpdate;
    //Stream.free;
    List.Free;
    end; {try-finally}
  except
  Invalidate;  // force repaint on exception
  raise;
  end; {try-except}
end;  {procedure}

procedure TzzCustomTreeView.SaveToFileEx(const FileName: string);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStreamEx(Stream);
  finally
    Fileclose(Stream.Handle);
    Stream.Free;
  end;
end;

procedure TzzCustomTreeView.LoadFromFileEx(const FileName:String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    FileClose(Stream.handle);
    Stream.Free;
  end;
end;

function TzzCustomTreeView.IsChecked(Node: TTreeNode): Boolean;
var
    TvItem  :TTVItem;
begin
    TvItem.Mask    :=  TVIF_STATE;
    TvItem.hItem  :=  Node.ItemId;
    TreeView_GetItem(Node.TreeView.Handle,  TvItem);
    Result := (TvItem.State  and  TVIS_CHECKED)  =  TVIS_CHECKED;
end;

procedure TzzCustomTreeView.SetChecked(Node: TTreeNode; Checked: Boolean);
var
    TvItem  :TTVItem;
begin
    FillChar(TvItem,  SizeOf(TvItem),  0);
    with TvItem do begin
      hItem := Node.ItemId;
      Mask := TVIF_STATE;
      StateMask := TVIS_STATEIMAGEMASK;
      if Checked then TvItem.State := TVIS_CHECKED
      else TvItem.State := TVIS_CHECKED  shr  1;
      TreeView_SetItem(Node.TreeView.Handle,  TvItem);
    end;
end;

function TzzCustomTreeView.CheckedCount: Integer;
var i: Integer;
begin
  Result := 0;
  for i := 0 to Pred(Items.Count) do
    if IsChecked(Items[i]) then
      Result := Result + 1;
end;

end.
